/*----------------------------------------------------------------------
    Find representative gray levels (by peak detection) findgray_pd.cxx
        Written by  H.Goto, Aug. 1999
        Modified by H.Goto, Sep. 1999
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1999  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


#define		MinCursorWidth	10

#define		GrayLevels	256
#define		MAXPEAKS	(GrayLevels /4)

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>

#include	"utypes.h"

#include	"siplib.h"
#include	"sgplib.h"
#include	"filters.h"
#include	"findcol_com.h"
#include	"findgray_pd.h"


static SIPImage	*edgeimage = 0;




/*----------------------------------------------
    Smoothing by moving average
----------------------------------------------*/

static int average7x(int *hist){
	int	i,j;
	int	sum;
	int	*tmphist;
	if ( 0 == (tmphist = new int[GrayLevels +6]) )  return(-1);
	memcpy((void *)(tmphist + 3),(void *)hist,sizeof(int) * GrayLevels);
	for ( i=0 ; i<3 ; i++ )  tmphist[i] = tmphist[3];
	for ( i=GrayLevels +5 ; i>=GrayLevels +3; i-- ) \
		tmphist[i] = tmphist[GrayLevels +2];
	for ( i=0 ; i<GrayLevels ; i++ ){
		sum = 0;
		for ( j=0 ; j<7 ; j++ )  sum += tmphist[i+j];
		hist[i] = sum;
	}
	delete []tmphist;
	return(0);
}




/*----------------------------------------------
    Get local gray level histogram
----------------------------------------------*/

static void get_localhist(SIPImage *image,int *hist, \
		int x0,int y0,int width,int height){
	int	x,y,sx,sy,ex,ey;
	uchar	*p;
	if ( 0 > (sx = x0) )  sx = 0;
	if ( 0 > (sy = y0) )  sy = 0;
	if ( image->width  < (ex = x0 + width ) )  ex = image->width ;
	if ( image->height < (ey = y0 + height) )  ey = image->height;
	for ( x=0 ; x<=GrayLevels ; x++ )  hist[x] = 0;
	for ( y=sy ; y<ey ; y++ ){
		p = (uchar *)sip_getimgptr(image,y);
		for ( x=sx ; x<ex ; x++ ){
			hist[ (int)((uint)p[x]) ]++;
		}
	}
	return;
}


static void get_localhist_weighted(SIPImage *image,SIPImage *edgeimage, \
		int *hist,int x0,int y0,int width,int height){
	int	x,y,sx,sy,ex,ey;
	uchar	*p,*wp;
	int	w;
	if ( 0 > (sx = x0) )  sx = 0;
	if ( 0 > (sy = y0) )  sy = 0;
	if ( image->width  < (ex = x0 + width ) )  ex = image->width ;
	if ( image->height < (ey = y0 + height) )  ey = image->height;
	for ( x=0 ; x<=GrayLevels ; x++ )  hist[x] = 0;
	for ( y=sy ; y<ey ; y++ ){
		p = (uchar *)sip_getimgptr(image,y);
		wp = (uchar *)sip_getimgptr(edgeimage,y);
		for ( x=sx ; x<ex ; x++ ){
			w = 64 - (int)((uint)wp[x]);
			if ( w < 0 )  w = 0;
			hist[ (int)((uint)p[x]) ] += w;
		}
	}
	for ( x=0 ; x<GrayLevels ; x++ )  hist[x] /= 64;
	return;
}




/*----------------------------------------------
    Prepare weight
----------------------------------------------*/

int findgray_pd_setweight(SIPImage *image){
	int	retcode;
	if ( edgeimage )  sip_DestroyImage(edgeimage);
	if ( 0 == (edgeimage = sip_CreateImage(image->width,image->height,8)) )  return(-1);
	retcode = edge_sobel(image,edgeimage,4);
//	xip_SaveImage("tmp.edge",edgeimage);
	return(retcode);
}




/*----------------------------------------------
    Find local representative gray levels
----------------------------------------------*/

int findgray_pd(SIPImage *image,int x0,int y0,int w,int h, \
		CCluster *ccluster,int maxclust,int wmode){
	int	n_cluster;
	int	i,c,l;
	int	hist[GrayLevels +1];
	int	cw = MinCursorWidth;	/* cursor width */
	peaklisti	peaklist[MAXPEAKS];

	if ( ! wmode ){
		get_localhist(image,hist,x0,y0,w,h);
	}
	else{	get_localhist_weighted(image,edgeimage,hist,x0,y0,w,h);
	}
	average7x(hist);

	/* hysteresis smoothing */
	c = 0;  l = 0;
	for ( i=0 ; i<GrayLevels ; i++ ){
		cw = (int)(0.2 * (double)hist[i]);
		if ( cw < MinCursorWidth )  cw = MinCursorWidth;

		/* force drop to background level for very long roof */
		if ( ++l > 16 ){
			c = hist[i];
			l = 0;
		}

		if ( hist[i] < c - cw ){
			c = hist[i] + cw;
			l = 0;
		}
		if ( hist[i] > c + cw ){
			c = hist[i] - cw;
			l = 0;
		}
		hist[i] = c;
	}

	/* peak detection */
	n_cluster = sgp_getpeaksi(hist,peaklist,\
		GrayLevels +1,MAXPEAKS,0,0x7fffffff);
	if ( n_cluster > maxclust )  n_cluster = maxclust;	/* limit */
	for ( i=0 ; i<n_cluster ; i++ ){
		ccluster[i].g = peaklist[i].pos;
		ccluster[i].count = peaklist[i].value;
	}
	return(n_cluster);
}




